@@ -3,9 +3,14 @@ var EXPORTED_SYMBOLS = ['cleanFilename', 'RemoteScript'];
 var Cc = Components.classes;
 var Ci = Components.interfaces;
 
+Components.utils.import("resource://greasemonkey/GM_notification.js");
+Components.utils.import('resource://greasemonkey/addons4.js');
 Components.utils.import('resource://greasemonkey/script.js');
+Components.utils.import('resource://greasemonkey/scriptIcon.js');
 Components.utils.import('resource://greasemonkey/util.js');
 
+Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
+
 var GM_config = GM_util.getService().config;
 var ioService = Cc['@mozilla.org/network/io-service;1']
     .getService(Ci.nsIIOService);
@@ -18,6 +23,14 @@ var stringBundleBrowser = Components
     .classes["@mozilla.org/intl/stringbundle;1"]
     .getService(Components.interfaces.nsIStringBundleService)
     .createBundle("chrome://greasemonkey/locale/gm-browser.properties");
+var xulRuntime = Components
+    .classes["@mozilla.org/xre/app-info;1"]
+    .getService(Components.interfaces.nsIXULRuntime);
+
+// See: http://goo.gl/vDSk9
+// Actual limit is 260; 240 ensures e.g. ".user.js" and slashes still fit.
+// The "/ 2" thing is so that we can have a directory, and a file in it.
+var windowsMaxNameLen = (240 - GM_util.scriptDir().path.length) / 2;
 
 /////////////////////////////// Private Helpers ////////////////////////////////
 
@@ -30,8 +43,24 @@ function cleanFilename(aFilename, aDefault) {
   // Blacklist problem characters (slashes, colons, etc.).
   var filename = (aFilename || aDefault).replace(
       disallowedFilenameCharacters, '');
+
   // Make whitespace readable.
   filename = filename.replace(/(\s|%20)+/g, '_');
+
+  // Limit length on Windows (#1548; http://goo.gl/vDSk9)
+  if ('WINNT' == xulRuntime.OS) {
+    if (windowsMaxNameLen <= 0) {
+      throw Error('Could not make a valid file name to save.');
+    }
+
+    var match = filename.match(/^(.+?)(\.(:?user\.js)|[^.{,8}])$/);
+    if (match) {
+      filename = match[1].substr(0, windowsMaxNameLen) + match[2];
+    } else {
+      filename = filename.substr(0, windowsMaxNameLen);
+    }
+  }
+
   // Ensure that it's something.
   if (!filename) filename = aDefault || 'unknown';
   return filename;
@@ -49,86 +78,149 @@ function filenameFromUri(aUri, aDefault) {
   return cleanFilename(filename, aDefault);
 }
 
-////////////////////////// Private Progress Listener ///////////////////////////
+////////////////////////// Private Download Listener ///////////////////////////
 
-function ProgressListener(
-    aRemoteScript, aUri, aCompletionCallback, aProgressCallback) {
-  this._completionCallback = aCompletionCallback || function() {};
-  this._progressCallback = aProgressCallback || function() {};
+function DownloadListener(
+    aTryToParse, aProgressCb, aCompletionCallback, aFile, aUri, aRemoteScript,
+    aErrorsAreFatal) {
+  this._completionCallback = aCompletionCallback;
+  this._data = [];
+  this._errorsAreFatal = ('undefined' == typeof aErrorsAreFatal)
+      ? true : aErrorsAreFatal;
+  this._progressCallback = aProgressCb;
   this._remoteScript = aRemoteScript;
+  this._tryToParse = aTryToParse;
   this._uri = aUri;
+
+  this._fileOutputStream = Cc["@mozilla.org/network/file-output-stream;1"]
+      .createInstance(Ci.nsIFileOutputStream);
+  this._fileOutputStream.init(aFile, -1, -1, null);
+  this._binOutputStream = Cc['@mozilla.org/binaryoutputstream;1']
+      .createInstance(Ci.nsIBinaryOutputStream);
+  this._binOutputStream.setOutputStream(this._fileOutputStream);
 }
 
-ProgressListener.prototype.onLocationChange = function(
-    aWebProgress, aRequest, aLocation) {
-};
+DownloadListener.prototype = {
+  _htmlTypeRegex: new RegExp('^text/(x|ht)ml', 'i'),
 
-ProgressListener.prototype.onProgressChange = function(
-    aWebProgress, aRequest, aCurSelfProgress, aMaxSelfProgress,
-    aCurTotalProgress, aMaxTotalProgress) {
-  var progress = aCurTotalProgress / aMaxTotalProgress;
-  if (-1 == aMaxTotalProgress) progress = 0;
-
-  if (this._progressCallback(aRequest, progress)) {
-    // The progress callback is where we check for HTML type, and return true
-    // (error status) if so.  In such a case, immediately complete as a failure.
-    this._completionCallback(aRequest, false, 'script');
-  }
-};
-
-ProgressListener.prototype.onSecurityChange = function(
-    aWebProgress, aRequest, aState) {
-};
-
-/** Called at least at the start and stop of the request. */
-ProgressListener.prototype.onStateChange = function(
-    aWebProgress, aRequest, aStateFlags, aStatus) {
-  // Find if there has been any error.
-  if (!(aStateFlags & Ci.nsIWebProgressListener.STATE_STOP)) {
-    return;
-  }
-  var error = aStatus !== 0;
-  var errorMessage = stringBundle.GetStringFromName('error.unknown');
-  try {
-    var httpChannel = aRequest.QueryInterface(Ci.nsIHttpChannel);
-    error |= !httpChannel.requestSucceeded;
-    error |= httpChannel.responseStatus >= 400;
-    errorMessage = stringBundle.GetStringFromName('error.serverReturned')
-        + ' ' + httpChannel.responseStatus + ' '
-        + httpChannel.responseStatusText + '.';
-  } catch (e) {
+  _parse: function(aRemoteScript) {
+    var converter = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
+        .createInstance(Ci.nsIScriptableUnicodeConverter);
+    converter.charset = 'UTF-8';
+    var source = '';
     try {
-      aRequest.QueryInterface(Ci.nsIFileChannel);
-      // no-op; if it got this far, aStatus is accurate.
+      source = converter.convertFromByteArray(this._data, this._data.length);
+    } catch (e) { }
+    return this._remoteScript.parseScript(source, false);
+  },
+
+  // nsIStreamListener
+  onDataAvailable: function(aRequest, aContext, aInputStream, aOffset, aCount) {
+    var binaryInputStream = Cc['@mozilla.org/binaryinputstream;1']
+        .createInstance(Ci.nsIBinaryInputStream);
+    binaryInputStream.setInputStream(aInputStream);
+
+    // Read incoming data.
+    var data = binaryInputStream.readByteArray(aCount);
+
+    if (this._tryToParse) {
+      // See #1823.  Strip UTF-8 BOM(s) at the very start of the file.
+      while (data && data.length >= 3
+          && data[0] == 0xEF && data[1] == 0xBB && data[2] == 0xBF
+          ) {
+        data = data.slice(3);
+      }
+
+      this._data = this._data.concat(data);
+      this._tryToParse = !this._parse(aContext);
+    } else {
+      this._data = null;
+    }
+
+    // Write it to the file.
+    this._binOutputStream.writeByteArray(data, data.length);
+  },
+
+  // nsIProgressEventSink
+  onProgress: function(aRequest, aContext, aProgress, aProgressMax) {
+    var progress;
+    if (-1 == aProgressMax || 0 == aProgressMax
+        || 0xFFFFFFFFFFFFFFFF == aProgressMax) {
+      progress = 0;
+    } else {
+      progress = aProgress / aProgressMax;
+    }
+    this._progressCallback(aRequest, progress);
+  },
+
+  // nsIRequestObserver
+  onStartRequest: function(aRequest, aContext) {
+    // For the first file (the script) detect an HTML page and abort if so.
+    if (this._tryToParse) {
+      try {
+        aRequest.QueryInterface(Ci.nsIHttpChannel);
+      } catch (e) {
+        return;  // Non-http channel?  Ignore.
+      }
+      if (this._htmlTypeRegex.test(aRequest.contentType)) {
+        this._completionCallback(aRequest, false, 'script');
+      }
+    }
+  },
+
+  // nsIRequestObserver
+  onStopRequest: function(aRequest, aContext, aStatusCode) {
+    this._binOutputStream.close();
+    this._fileOutputStream.close();
+
+    var error = aStatusCode !== 0;
+    var errorMessage = stringBundle.GetStringFromName('error.unknown');
+    try {
+      var httpChannel = aRequest.QueryInterface(Ci.nsIHttpChannel);
+      error |= !httpChannel.requestSucceeded;
+      error |= httpChannel.responseStatus >= 400;
+      errorMessage = stringBundle.GetStringFromName('error.serverReturned')
+          + ' ' + httpChannel.responseStatus + ' '
+          + httpChannel.responseStatusText + '.';
     } catch (e) {
-      dump('aRequest is neither http nor file channel: ' + aRequest + '\n');
-      for (i in Ci) {
-        try {
-          aRequest.QueryInterface(Ci[i]);
-          dump('it is a: ' + i + '\n');
-        } catch (e) {
-          // ignore
+      try {
+        aRequest.QueryInterface(Ci.nsIFileChannel);
+        // no-op; if it got this far, aStatus is accurate.
+      } catch (e) {
+        dump('aRequest is neither http nor file channel: ' + aRequest + '\n');
+        for (i in Ci) {
+          try {
+            aRequest.QueryInterface(Ci[i]);
+            dump('it is a: ' + i + '\n');
+          } catch (e) {
+            // ignore
+          }
         }
       }
     }
-  }
 
-  if (error) {
-    errorMessage = stringBundle.GetStringFromName('error.downloadingUrl')
-        + '\n' + this._uri.spec + '\n\n' + errorMessage;
-    this._remoteScript.cleanup(errorMessage);
-  }
-  error |= this._progressCallback(aRequest, 1);
-  this._completionCallback(aRequest, !error);
-};
+    if (error && this._errorsAreFatal) {
+      errorMessage = stringBundle.GetStringFromName('error.downloadingUrl')
+          + '\n' + this._uri.spec + '\n\n' + errorMessage;
+      this._remoteScript.cleanup(errorMessage);
+    }
 
-ProgressListener.prototype.onStatusChange = function(
-    aWebProgress, aRequest, aStatus, aMessage) {
-  // TODO: Better figure out the possible aStatus values.  Docs say:
-  // "This interface does not define the set of possible status codes."
+    this._progressCallback(aRequest, 1);
+    this._completionCallback(aRequest, !error);
+  },
 
-  // Manually found when reading an invalid file:/// URL.
-  if (2152857618 == aStatus) this._completionCallback(aRequest, false);
+  // nsIProgressEventSink
+  onStatus: function(aRequest, aContext, aStatus, aStatusArg) { },
+
+  // nsIInterfaceRequestor
+  getInterface: function(aIID) { return this.QueryInterface(aIID); },
+
+  // nsISupports
+  QueryInterface: XPCOMUtils.generateQI([
+      Ci.nsIProgressEventSink,
+      Ci.nsIStreamListener,
+      Ci.nsISupports,
+      ]),
 };
 
 /////////////////////////////// Public Interface ///////////////////////////////
@@ -138,6 +230,7 @@ ProgressListener.prototype.onStatusChange = function(
 
 function RemoteScript(aUrl) {
   this._baseName = null;
+  this._channels = [];
   this._dependencies = [];
   this._metadata = null;
   this._progress = [0, 0];
@@ -145,12 +238,14 @@ function RemoteScript(aUrl) {
   this._progressIndex = 0;
   this._scriptFile = null;
   this._scriptMetaCallbacks = [];
+  this._silent = false;
   this._tempDir = GM_util.getTempDir();
   this._uri = GM_util.uriFromUrl(aUrl);
   this._url = aUrl;
 
   this.done = false;
   this.errorMessage = null;
+  this.messageName = 'script.installed';
   this.script = null;
 }
 
@@ -162,7 +257,14 @@ RemoteScript.prototype.cleanup = function(aErrorMessage) {
   this.errorMessage = aErrorMessage || null;
   this.done = true;
 
-  if (this._wbp) this._wbp.cancelSave();
+  this._channels.forEach(function(aChannel) {
+    try {
+      aChannel.QueryInterface(Ci.nsIRequest);
+    } catch (e) {
+      return;
+    }
+    aChannel.cancel(Components.results.NS_BINDING_ABORTED);
+  });
   if (this._tempDir && this._tempDir.exists()) {
     this._tempDir.remove(true);
   }
@@ -177,8 +279,6 @@ RemoteScript.prototype.download = function(aCompletionCallback) {
       aCompletionCallback, 'Completion callback is not a function.');
 
   if (this.script) {
-    // TODO: Verify that this condition really is sufficient.  Is the script
-    // completely loaded?
     this._downloadDependencies(aCompletionCallback);
   } else {
     this.downloadScript(GM_util.hitch(this, function(aSuccess, aPoint) {
@@ -203,12 +303,15 @@ RemoteScript.prototype.downloadScript = function(aCompletionCallback) {
       this._tempDir, filenameFromUri(this._uri, 'gm_script'));
 
   this._downloadFile(this._uri, this._scriptFile,
-      GM_util.hitch(this, this._downloadScriptCb, aCompletionCallback));
+      GM_util.hitch(this, this._downloadScriptCb, aCompletionCallback),
+      true // aErrorsAreFatal
+      );
 };
 
 RemoteScript.prototype.install = function(aOldScript, aOnlyDependencies) {
   if (!this.script) {
-    throw new Error('RemoteScript.install(): Script is not downloaded.');
+    throw new Error(
+        stringBundle.GetStringFromName('remotescript.not-downloaded'));
   }
   if ('undefined' == typeof aOnlyDependencies) aOnlyDependencies = false;
 
@@ -216,17 +319,17 @@ RemoteScript.prototype.install = function(aOldScript, aOnlyDependencies) {
     // Just move the dependencies in.
     var enumerator = this._tempDir.directoryEntries;
     while (enumerator.hasMoreElements()) {
-      var file = enumerator.getNext().QueryInterface(Ci.nsILocalFile);
-      // TODO: Fix invalid private access.
-      file.moveTo(this.script._basedirFile, null);
+      var file = enumerator.getNext().QueryInterface(Ci.nsIFile);
+      file.moveTo(this.script.baseDirFile, null);
     }
   } else {
     // Completely install the new script.
     if (!this._baseName) {
-      throw new Error('RemoteScript.install(): Script base name unknown.');
+      throw new Error(
+          stringBundle.GetStringFromName('remotescript.name-unknown'));
     }
 
-    GM_config.install(this.script, aOldScript);
+    GM_config.install(this.script, aOldScript, this._tempDir);
 
     var suffix = 0;
     var file = GM_util.scriptDir();
@@ -241,6 +344,22 @@ RemoteScript.prototype.install = function(aOldScript, aOnlyDependencies) {
     this.script.setFilename(this._baseName, this._scriptFile.leafName);
     this._tempDir.moveTo(GM_util.scriptDir(), this._baseName);
     this._tempDir = null;
+
+    this.script.fixTimestampsOnInstall();
+    this.script.checkConfig();
+
+    // Now that we've fully populated the new state, update the AOM and config
+    // data based on that.
+    ScriptAddonFactoryByScript(this.script, true);
+    this.script._changed('modified', this.script.id);
+
+    // Let the user know we're all done.
+    if (!this._silent) {
+      GM_notification(
+          "'" + this.script.name + "' "
+              + stringBundleBrowser.GetStringFromName(this.messageName),
+          this.messageName);
+    }
   }
 };
 
@@ -256,6 +375,31 @@ RemoteScript.prototype.onScriptMeta = function(aCallback) {
   this._scriptMetaCallbacks.push(aCallback);
 };
 
+/** Parse the source code of the script, discover dependencies, data & etc. */
+RemoteScript.prototype.parseScript = function(aSource, aFatal) {
+  if (this.errorMessage) return false;
+  if (this.script) return true;
+
+  var scope = {};
+  Components.utils.import('resource://greasemonkey/parseScript.js', scope);
+  var script = scope.parse(aSource, this._uri, true, aFatal);
+  if (!script || script.parseErrors.length) {
+    if (aFatal) {
+      this.cleanup(
+          stringBundle.GetStringFromName('error.parsingScript')
+          + '\n' + script.parseErrors);
+    }
+    return false;
+  }
+
+  this._baseName = cleanFilename(script.name, 'gm-script');
+  this._dispatchCallbacks('scriptMeta', script);
+  this.script = script;
+  this._postParseScript();
+
+  return true;
+};
+
 /** Set the (installed) script, in order to download modified dependencies.
  *
  * After calling this, calling .download() will only get dependencies.  This
@@ -270,7 +414,11 @@ RemoteScript.prototype.setScript = function(aScript, aTempFile) {
     this._scriptFile = aTempFile;
     this._baseName = cleanFilename(aScript.name, 'gm-script');
   }
-  this._postParseScriptFile();
+  this._postParseScript();
+};
+
+RemoteScript.prototype.setSilent = function(aVal) {
+  this._silent = !!aVal;
 };
 
 RemoteScript.prototype.showSource = function(aTabBrowser) {
@@ -284,28 +432,24 @@ RemoteScript.prototype.showSource = function(aTabBrowser) {
   var tab = aTabBrowser.loadOneTab(
       ioService.newFileURI(this._scriptFile).spec,
       {'inBackground': false});
-  function addNotification() {
-    var notificationBox = aTabBrowser.getNotificationBox();
-    notificationBox.appendNotification(
-      stringBundleBrowser.GetStringFromName('greeting.msg'),
-      "install-userscript",
-      "chrome://greasemonkey/skin/icon16.png",
-      notificationBox.PRIORITY_WARNING_MEDIUM,
-      [{
-        'label': stringBundleBrowser.GetStringFromName('greeting.btn'),
-        'accessKey': stringBundleBrowser.GetStringFromName('greeting.btnAccess'),
-        'popup': null,
-        'callback': GM_util.hitch(this, function() {
-          GM_util.showInstallDialog(this, aTabBrowser, GM_util.getService());
-          // Timeout puts this after the notification closes itself for the
-          // button click, avoiding an error inside that (Firefox) code.
-          GM_util.timeout(0, function() { aTabBrowser.removeTab(tab); });
-        })
-      }]
-    );
-  }
-  // Timeout necessary only for Firefox 3 as a dirty hack around timing issues.
-  GM_util.timeout(0, GM_util.hitch(this, addNotification));
+  var notificationBox = aTabBrowser.getNotificationBox();
+  notificationBox.appendNotification(
+    stringBundleBrowser.GetStringFromName('greeting.msg'),
+    "install-userscript",
+    "chrome://greasemonkey/skin/icon16.png",
+    notificationBox.PRIORITY_WARNING_MEDIUM,
+    [{
+      'label': stringBundleBrowser.GetStringFromName('greeting.btn'),
+      'accessKey': stringBundleBrowser.GetStringFromName('greeting.btnAccess'),
+      'popup': null,
+      'callback': GM_util.hitch(this, function() {
+        GM_util.showInstallDialog(this, aTabBrowser, GM_util.getService());
+        // Timeout puts this after the notification closes itself for the
+        // button click, avoiding an error inside that (Firefox) code.
+        GM_util.timeout(function() { aTabBrowser.removeTab(tab); }, 0);
+      })
+    }]
+  );
 };
 
 RemoteScript.prototype.toString = function() {
@@ -338,14 +482,18 @@ RemoteScript.prototype._downloadDependencies = function(aCompletionCallback) {
   // Because _progressIndex includes the base script at 0, subtract one to
   // get the dependency index.
   var dependency = this._dependencies[this._progressIndex - 1];
-  var uri = GM_util.uriFromUrl(dependency.urlToDownload);
+  var uri = GM_util.uriFromUrl(dependency.downloadURL);
   var file = GM_util.getTempFile(this._tempDir, filenameFromUri(uri));
   dependency.setFilename(file);
 
   function dependencyDownloadComplete(aChannel, aSuccess) {
     if (!aSuccess) {
-      aCompletionCallback(aSuccess, 'dependency');
-      return;
+      if (dependency instanceof ScriptIcon) {
+        // Ignore the failure to download the icon.
+      } else {
+        aCompletionCallback(aSuccess, 'dependency');
+        return;
+      }
     }
     if (dependency.setMimetype) {
       dependency.setMimetype(aChannel.contentType);
@@ -357,14 +505,16 @@ RemoteScript.prototype._downloadDependencies = function(aCompletionCallback) {
   }
 
   this._downloadFile(
-      uri, file, GM_util.hitch(this, dependencyDownloadComplete));
+      uri, file, GM_util.hitch(this, dependencyDownloadComplete),
+      !(dependency instanceof ScriptIcon)  // aErrorsAreFatal
+      );
 };
 
-/** Download a given nsIURI to a given nsILocalFile, with optional callback. */
+/** Download a given nsIURI to a given nsIFile, with optional callback. */
 RemoteScript.prototype._downloadFile = function(
-    aUri, aFile, aCompletionCallback) {
+    aUri, aFile, aCompletionCallback, aErrorsAreFatal) {
   aUri = aUri.QueryInterface(Ci.nsIURI);
-  aFile = aFile.QueryInterface(Ci.nsILocalFile);
+  aFile = aFile.QueryInterface(Ci.nsIFile);
   aCompletionCallback = aCompletionCallback || function() {};
   assertIsFunction(aCompletionCallback,
       '_downloadFile() completion callback is not a function.');
@@ -377,66 +527,61 @@ RemoteScript.prototype._downloadFile = function(
       // No-op, always allow files from the same scheme as the script.
     } else if (!GM_util.isGreasemonkeyable(aUri.spec)) {
       // Otherwise, these are unsafe.  Do not download them.
-      this.cleanup('Will not download unsafe URL:\n' + aUri.spec);
+      this.cleanup(
+          stringBundle.GetStringFromName('remotescript.unsafe-url')
+              .replace('%1', aUri.spec)
+          );
       return;
     }
   }
 
-  // Dangerous semi-global state:  The web browser persist object is stored
-  // in the object, so that it can be canceled.  Parallel downloads would need
-  // to be handled differently.
   var channel = ioService.newChannelFromURI(aUri);
-  this._wbp = Cc['@mozilla.org/embedding/browser/nsWebBrowserPersist;1']
-      .createInstance(Ci.nsIWebBrowserPersist);
-  this._wbp.persistFlags =
-      Ci.nsIWebBrowserPersist.PERSIST_FLAGS_BYPASS_CACHE |
-      Ci.nsIWebBrowserPersist.PERSIST_FLAGS_REPLACE_EXISTING_FILES |
-      Ci.nsIWebBrowserPersist.PERSIST_FLAGS_CLEANUP_ON_FAILURE |
-      Ci.nsIWebBrowserPersist.PERSIST_FLAGS_FORCE_ALLOW_COOKIES;
-  this._wbp.progressListener = new ProgressListener(
-      this, aUri,
-      aCompletionCallback, GM_util.hitch(this, this._downloadFileProgress));
-  this._wbp.saveChannel(channel, aFile);
+  this._channels.push(channel);
+  var dsl = new DownloadListener(
+      0 == this._progressIndex,  // aTryToParse
+      GM_util.hitch(this, this._downloadFileProgress),
+      aCompletionCallback,
+      aFile,
+      aUri,
+      this,
+      aErrorsAreFatal
+      );
+  channel.notificationCallbacks = dsl;
+  channel.asyncOpen(dsl, this);
 };
 
-RemoteScript.prototype._htmlTypeRegex = new RegExp('^text/(x|ht)ml', 'i');
 RemoteScript.prototype._downloadFileProgress = function(
     aChannel, aFileProgress) {
-  if (0 == this._progressIndex && !this.script) {
-    // We are downloading the first file, and haven't parsed a script yet ...
-
-    // 1) Detect an HTML page and abort if so.
-    if (this._htmlTypeRegex.test(aChannel.contentType)) {
-      this.cleanup();
-      return true;
-    }
-
-    // 2) Otherwise try to parse the script from the downloaded file.
-    this.script = this._parseScriptFile();
-    if (this.script) {
-      // And if successful, prepare to download dependencies.
-      this._postParseScriptFile();
-    }
-  }
-
   this._progress[this._progressIndex] = aFileProgress;
   var progress = this._progress.reduce(function(a, b) { return a + b; })
       / this._progress.length;
-
   this._dispatchCallbacks('progress', progress);
-
-  return false;
 };
 
 RemoteScript.prototype._downloadScriptCb = function(
     aCompletionCallback, aChannel, aSuccess) {
   if (aSuccess) {
     // At this point downloading the script itself is definitely done.
-    if (!this.script) {
-      // If we don't have a script object, we failed to find metadata and parse
-      // it during download.  Try one last time.
-      this.script = this._parseScriptFile(true);
+
+    // Parse the script.
+    try {
+      this._parseScriptFile();
+    } catch (e) {
+      // If that failed, set the error message, and ...
+      if (-1 === new String(e).indexOf('Unicode')) {
+        this.cleanup(stringBundle.GetStringFromName('error.unknown'));
+      } else {
+        this.cleanup(stringBundle.GetStringFromName('error.scriptCharset'));
+      }
     }
+
+    if (this.errorMessage) {
+      // Fake a successful download, so the install window will show, with
+      // the error message.
+      this._dispatchCallbacks('scriptMeta', new Script());
+      return aCompletionCallback(true, 'script');
+    }
+
     if (!this.script) {
       dump('RemoteScript: finishing with error because no script was found.\n');
       // If we STILL don't have a script, this is a fatal error.
@@ -448,29 +593,20 @@ RemoteScript.prototype._downloadScriptCb = function(
   aCompletionCallback(aSuccess, 'script');
 };
 
-/** Produce a Script object from the contents of this._scriptFile. */
 RemoteScript.prototype._parseScriptFile = function() {
-  if (this.errorMessage) return;
-
-  var source = GM_util.getContents(this._scriptFile);
+  if (this.done) return;
+  var source = GM_util.getContents(this._scriptFile, null, true);
   if (!source) return null;
-
-  var scope = {};
-  Components.utils.import('resource://greasemonkey/parseScript.js', scope);
-  var script = scope.parse(source, this._uri);
-  if (script.parseErrors.length) {
-    this.cleanup(
-        stringBundle.GetStringFromName('error.parsingScript')
-        + '\n' + script.parseErrors);
-    return null;
+  var script = null;
+  try {
+    this.parseScript(source, true);
+  } catch (e) {
+    dump('RemoteScript._parseScriptFile: ' + e + '\n');
   }
-  this._baseName = cleanFilename(script.name, 'gm-script');
-  this._dispatchCallbacks('scriptMeta', script);
-
   return script;
 };
 
-RemoteScript.prototype._postParseScriptFile = function() {
+RemoteScript.prototype._postParseScript = function() {
   this._dependencies = this.script.dependencies;
   this._progress = [];
   for (var i = 0; i < this._dependencies.length; i++) {
