@@ -2,10 +2,10 @@ Components.utils.import('resource://greasemonkey/constants.js');
 
 const EXPORTED_SYMBOLS = ['getTempDir'];
 
-const DIRECTORY_TYPE = Components.interfaces.nsILocalFile.DIRECTORY_TYPE;
+const DIRECTORY_TYPE = Components.interfaces.nsIFile.DIRECTORY_TYPE;
 const TMP_DIR = Components.classes["@mozilla.org/file/directory_service;1"]
     .getService(Components.interfaces.nsIProperties)
-    .get("TmpD", Components.interfaces.nsILocalFile);
+    .get("TmpD", Components.interfaces.nsIFile);
 
 function getTempDir(aRoot) {
   var file = (aRoot || TMP_DIR).clone();
